import pytest

from canals.component import component
from canals.errors import ComponentError


def test_correct_declaration():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: int

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    # Verifies also instantiation works with no issues
    assert MockComponent()
    assert component.registry["MockComponent"] == MockComponent


def test_input_required():
    with pytest.raises(
        ComponentError,
        match="No input definition found in Component MockComponent. "
        "Create a method that returns a dataclass defining the input and "
        "decorate it with @component.input\(\) to fix the error.",
    ):

        @component
        class MockComponent:
            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output

            def run(self, data):
                return MockComponent.Output(output_value=1)


def test_output_required():
    with pytest.raises(
        ComponentError,
        match="No output definition found in Component MockComponent. "
        "Create a method that returns a dataclass defining the output and "
        "decorate it with @component.output\(\) to fix the error.",
    ):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            def run(self, data):
                return 1


def test_only_single_input_defined():
    with pytest.raises(
        ComponentError,
        match="Multiple input definitions found for Component MockComponent",
    ):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            @component.input
            def another_input(self):
                class Input:
                    input_value: int

                return Input

            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output

            def run(self, data):
                return self.output(output_value=1)


def test_only_single_output_defined():
    with pytest.raises(
        ComponentError,
        match="Multiple output definitions found for Component MockComponent",
    ):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output

            @component.output
            def another_output(self):
                class Output:
                    output_value: int

                return Output

            def run(self, data):
                return self.output(output_value=1)


def test_check_for_run():
    with pytest.raises(ComponentError, match="must have a 'run\(\)' method"):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output


def test_run_takes_only_one_param():
    with pytest.raises(ComponentError, match="must accept only a single parameter called 'data'."):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output

            def run(self, data, something_else: int):
                return self.output(output_value=1)


def test_run_takes_only_kwarg_data():
    with pytest.raises(ComponentError, match="must accept a parameter called 'data'."):

        @component
        class MockComponent:
            @component.input
            def input(self):
                class Input:
                    input_value: int

                return Input

            @component.output
            def output(self):
                class Output:
                    output_value: int

                return Output

            def run(self, wrong_name):
                return self.output(output_value=1)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Any, Dict, List, Literal, Union

import os
import json
import datetime
import logging
from pathlib import Path
from copy import deepcopy
from collections import OrderedDict
from dataclasses import fields

import networkx

from canals.component import Component
from canals.errors import PipelineConnectError, PipelineMaxLoops, PipelineRuntimeError, PipelineValidationError
from canals.draw import draw, convert_for_debug, RenderingEngines
from canals.pipeline.sockets import InputSocket, OutputSocket, find_input_sockets, find_output_sockets
from canals.pipeline.validation import validate_pipeline_input
from canals.pipeline.connections import parse_connection_name, find_unambiguous_connection, get_socket_type_desc


logger = logging.getLogger(__name__)


class Pipeline:
    """
    Components orchestration engine.

    Builds a graph of components and orchestrates their execution according to the execution graph.
    """

    def __init__(
        self,
        metadata: Optional[Dict[str, Any]] = None,
        max_loops_allowed: int = 100,
        debug_path: Union[Path, str] = Path(".canals_debug/"),
    ):
        """
        Creates the Pipeline.

        Args:
            metadata: arbitrary dictionary to store metadata about this pipeline. Make sure all the values contained in
                this dictionary can be serialized and deserialized if you wish to save this pipeline to file with
                `save_pipelines()/load_pipelines()`.
            max_loops_allowed: how many times the pipeline can run the same node before throwing an exception.
            debug_path: when debug is enabled in `run()`, where to save the debug data.
        """
        self.metadata = metadata or {}
        self.max_loops_allowed = max_loops_allowed
        self.graph = networkx.MultiDiGraph()
        self.debug: Dict[int, Dict[str, Any]] = {}
        self.debug_path = Path(debug_path)

    def __eq__(self, other) -> bool:
        """
        Equal pipelines share every metadata, node and edge, but they're not required to use
        the same node instances: this allows pipeline saved and then loaded back to be equal to themselves.
        """
        if (
            not isinstance(other, type(self))
            or not getattr(self, "metadata") == getattr(other, "metadata")
            or not getattr(self, "max_loops_allowed") == getattr(other, "max_loops_allowed")
            or not hasattr(self, "graph")
            or not hasattr(other, "graph")
        ):
            return False

        return (
            self.graph.adj == other.graph.adj
            and self._comparable_nodes_list(self.graph) == self._comparable_nodes_list(other.graph)
            and self.graph.graph == other.graph.graph
        )

    def _comparable_nodes_list(self, graph: networkx.MultiDiGraph) -> List[Dict[str, Any]]:
        """
        Replaces instances of nodes with their class name and defaults list in order to make sure they're comparable.
        """
        nodes = []
        for node in graph.nodes:
            comparable_node = graph.nodes[node]
            if hasattr(comparable_node, "defaults"):
                comparable_node["defaults"] = comparable_node["instance"].defaults
            comparable_node["instance"] = comparable_node["instance"].__class__
            nodes.append(comparable_node)
        nodes.sort()
        return nodes

    def add_component(self, name: str, instance: Component) -> None:
        """
        Create a component for the given component. Components are not connected to anything by default:
        use `Pipeline.connect()` to connect components together.

        Component names must be unique, but component instances can be reused if needed.

        Args:
            name: the name of the component.
            instance: the component instance.

        Returns:
            None

        Raises:
            ValueError: if a component with the same name already exists
            PipelineValidationError: if the given instance is not a Canals component
        """
        # Component names are unique
        if name in self.graph.nodes:
            raise ValueError(f"A component named '{name}' already exists in this pipeline: choose another name.")

        # Components can't be named `_debug`
        if name == "_debug":
            raise ValueError("'_debug' is a reserved name for debug output. Choose another name.")

        # Component instances must be components
        if not hasattr(instance, "__canals_component__"):
            raise PipelineValidationError(
                f"'{type(instance)}' doesn't seem to be a component. Is this class decorated with @component?"
            )

        # Find inputs and outputs
        input_sockets = find_input_sockets(instance)
        output_sockets = find_output_sockets(instance)

        # Add component to the graph, disconnected
        logger.debug("Adding component '%s' (%s)", name, instance)
        self.graph.add_node(
            name,
            instance=instance,
            input_sockets=input_sockets,
            output_sockets=output_sockets,
            visits=0,
        )

    def connect(self, connect_from: str, connect_to: str) -> None:
        """
        Connects two components together. All components to connect must exist in the pipeline.
        If connecting to an component that has several output connections, specify the inputs and output names as
        'component_name.connections_name'.

        Args:
            connect_from: the component that delivers the value. This can be either just a component name or can be
                in the format `component_name.connection_name` if the component has multiple outputs.
            connect_to: the component that receives the value. This can be either just a component name or can be
                in the format `component_name.connection_name` if the component has multiple inputs.

        Returns:
            None

        Raises:
            PipelineConnectError: if the two components cannot be connected (for example if one of the components is
                not present in the pipeline, or the connections don't match by type, and so on).
        """
        # Edges may be named explicitly by passing 'node_name.edge_name' to connect().
        from_node, from_socket_name = parse_connection_name(connect_from)
        to_node, to_socket_name = parse_connection_name(connect_to)

        # Get the nodes data.
        try:
            from_sockets = self.graph.nodes[from_node]["output_sockets"]
        except KeyError as exc:
            raise ValueError(f"Component named {from_node} not found in the pipeline.") from exc

        try:
            to_sockets = self.graph.nodes[to_node]["input_sockets"]
        except KeyError as exc:
            raise ValueError(f"Component named {to_node} not found in the pipeline.") from exc

        # If the name of either socket is given, get the socket
        if from_socket_name:
            from_socket = from_sockets.get(from_socket_name, None)
            if not from_socket:
                raise PipelineConnectError(
                    f"'{from_node}.{from_socket_name} does not exist. "
                    f"Output connections of {from_node} are: "
                    + ", ".join(
                        [f"{name} (type {get_socket_type_desc(socket.type)})" for name, socket in from_sockets.items()]
                    )
                )
        if to_socket_name:
            to_socket = to_sockets.get(to_socket_name, None)
            if not to_socket:
                raise PipelineConnectError(
                    f"'{to_node}.{to_socket_name} does not exist. "
                    f"Input connections of {to_node} are: "
                    + ", ".join(
                        [f"{name} (type {get_socket_type_desc(socket.type)})" for name, socket in to_sockets.items()]
                    )
                )

        # Look for an unambiguous connection among the possible ones.
        # Note that if there is more than one possible connection but two sockets match by name, they're paired.
        from_sockets = [from_socket] if from_socket_name else list(from_sockets.values())
        to_sockets = [to_socket] if to_socket_name else list(to_sockets.values())
        from_socket, to_socket = find_unambiguous_connection(
            sender_node=from_node, sender_sockets=from_sockets, receiver_node=to_node, receiver_sockets=to_sockets
        )

        # Connect the components on these sockets
        self._direct_connect(from_node=from_node, from_socket=from_socket, to_node=to_node, to_socket=to_socket)

    def _direct_connect(self, from_node: str, from_socket: OutputSocket, to_node: str, to_socket: InputSocket) -> None:
        """
        Directly connect socket to socket. This method does not type-check the connections: use 'Pipeline.connect()'
        instead (which uses 'find_unambiguous_connection()' to validate types).
        """
        # Make sure the receiving socket isn't already connected - sending sockets can be connected as many times as needed,
        # so they don't need this check
        if to_socket.sender:
            raise PipelineConnectError(
                f"Cannot connect '{from_node}.{from_socket.name}' with '{to_node}.{to_socket.name}': "
                f"{to_node}.{to_socket.name} is already connected to {to_socket.sender}.\n"
            )

        # Create the connection
        logger.debug("Connecting '%s.%s' to '%s.%s'", from_node, from_socket.name, to_node, to_socket.name)
        edge_key = f"{from_socket.name}/{to_socket.name}"
        self.graph.add_edge(from_node, to_node, key=edge_key, from_socket=from_socket, to_socket=to_socket)

        # Stores the name of the node that will send its output to this socket
        to_socket.sender = from_node

    def get_component(self, name: str) -> Component:
        """
        Returns an instance of a component.

        Args:
            name: the name of the component

        Returns:
            The instance of that component.

        Raises:
            ValueError: if a component with that name is not present in the pipeline.
        """
        try:
            return self.graph.nodes[name]["instance"]
        except KeyError as exc:
            raise ValueError(f"Component named {name} not found in the pipeline.") from exc

    def draw(self, path: Path, engine: RenderingEngines = "mermaid-img") -> None:
        """
        Draws the pipeline. Requires either `graphviz` as a system dependency, or an internet connection for Mermaid.
        Run `pip install canals[graphviz]` or `pip install canals[mermaid]` to install missing dependencies.

        Args:
            path: where to save the diagram.
            engine: which format to save the graph as. Accepts 'graphviz', 'mermaid-text', 'mermaid-img'.
                Default is 'mermaid-img'.

        Returns:
            None

        Raises:
            ImportError: if `engine='graphviz'` and `pygraphviz` is not installed.
            HTTPConnectionError: (and similar) if the internet connection is down or other connection issues.
        """
        draw(graph=deepcopy(self.graph), path=path, engine=engine)

    def warm_up(self):
        """
        Make sure all nodes are warm.

        It's the node's responsibility to make sure this method can be called at every `Pipeline.run()`
        without re-initializing everything.
        """
        for node in self.graph.nodes:
            if hasattr(self.graph.nodes[node]["instance"], "warm_up"):
                logger.info("Warming up component %s...", node)
                self.graph.nodes[node]["instance"].warm_up()

    def run(self, data: Dict[str, Any], debug: bool = False) -> Dict[str, Any]:
        """
        Runs the pipeline.

        Args:
            data: the inputs to give to the input components of the Pipeline.
            parameters: a dictionary with all the parameters of all the components, namespaced by component.
            debug: whether to collect and return debug information.

        Returns:
            A dictionary with the outputs of the output components of the Pipeline.

        Raises:
            PipelineRuntimeError: if the any of the components fail or return unexpected output.
        """
        # **** The Pipeline.run() algorithm ****
        #
        # Nodes are run as soon as an input for them appears in the inputs buffer.
        # When there's more than a node at once in the buffer (which means some
        # branches are running in parallel or that there are loops) they are selected to
        # run in FIFO order by the `inputs_buffer` OrderedDict.
        #
        # Inputs are labeled with the name of the node they're aimed for:
        #
        #   ````
        #   inputs_buffer[target_node] = {"input_name": input_value, ...}
        #   ```
        #
        # Nodes should wait until all the necessary input data has arrived before running.
        # If they're popped from the input_buffer before they're ready, they're put back in.
        # If the pipeline has branches of different lengths, it's possible that a node has to
        # wait a bit and let other nodes pass before receiving all the input data it needs.
        #
        # Chetsheet for networkx data access:
        # - Name of the node       # self.graph.nodes  (List[str])
        # - Node instance          # self.graph.nodes[node]["instance"]
        # - Input nodes            # [e[0] for e in self.graph.in_edges(node)]
        # - Output nodes           # [e[1] for e in self.graph.out_edges(node)]
        # - Output edges           # [e[2]["label"] for e in self.graph.out_edges(node, data=True)]
        #
        # if debug:
        #     os.makedirs("debug", exist_ok=True)

        data = validate_pipeline_input(self.graph, input_values=data)
        self._clear_visits_count()
        self.warm_up()

        logger.info("Pipeline execution started.")
        inputs_buffer = OrderedDict(
            {
                node: {key: value for key, value in input_data.__dict__.items() if value is not None}
                for node, input_data in data.items()
            }
        )
        pipeline_output: Dict[str, Dict[str, Any]] = {}

        if debug:
            logger.info("Debug mode ON.")
        self.debug = {}

        # *** PIPELINE EXECUTION LOOP ***
        # We select the nodes to run by popping them in FIFO order from the inputs buffer.
        step = 0
        while inputs_buffer:
            step += 1
            if debug:
                self._record_pipeline_step(step, inputs_buffer, pipeline_output)
            logger.debug("> Queue at step %s: %s", step, {k: list(v.keys()) for k, v in inputs_buffer.items()})

            component, inputs = inputs_buffer.popitem(last=False)  # FIFO

            # Make sure it didn't run too many times already
            self._check_max_loops(component)

            # **** IS IT MY TURN YET? ****
            # Check if the component should be run or not
            action = self._calculate_action(name=component, inputs=inputs, inputs_buffer=inputs_buffer)

            # This component is missing data: let's put it back in the queue and wait.
            if action == "wait":
                if not inputs_buffer:
                    # What if there are no components to wait for?
                    raise PipelineRuntimeError(
                        f"'{component}' is stuck waiting for input, but there are no other components to run. "
                        "This is likely a Canals bug. Open an issue at https://github.com/deepset-ai/canals."
                    )

                inputs_buffer[component] = inputs
                continue

            # This component did not receive the input it needs: it must be on a skipped branch. Let's not run it.
            if action == "skip":
                self.graph.nodes[component]["visits"] += 1
                inputs_buffer = self._skip_downstream_unvisited_nodes(component=component, inputs_buffer=inputs_buffer)
                continue

            if action == "remove":
                # This component has no reason of being in the run queue and we need to remove it. For example, this can happen to components that are connected to skipped branches of the pipeline.
                continue

            # **** RUN THE NODE ****
            # It is our turn! The node is ready to run and all necessary inputs are present
            output = self._run_component(name=component, inputs=inputs)

            # **** PROCESS THE OUTPUT ****
            # The node run successfully. Let's store or distribute the output it produced, if it's valid.
            if not self.graph.out_edges(component):
                # Note: if a node outputs many times (like in loops), the output will be overwritten
                pipeline_output[component] = output
            else:
                inputs_buffer = self._route_output(
                    node_results=output, node_name=component, inputs_buffer=inputs_buffer
                )

        if debug:
            self._record_pipeline_step(step + 1, inputs_buffer, pipeline_output)

            # Save to json
            os.makedirs(self.debug_path, exist_ok=True)
            with open(self.debug_path / "data.json", "w", encoding="utf-8") as datafile:
                json.dump(self.debug, datafile, indent=4, default=str)

            # Store in the output
            pipeline_output["_debug"] = self.debug  # type: ignore

        logger.info("Pipeline executed successfully.")
        return pipeline_output

    def _record_pipeline_step(self, step, inputs_buffer, pipeline_output):
        """
        Stores a snapshot of this step into the self.debug dictionary of the pipeline.
        """
        mermaid_graph = convert_for_debug(deepcopy(self.graph))
        self.debug[step] = {
            "time": datetime.datetime.now(),
            "inputs_buffer": list(inputs_buffer.items()),
            "pipeline_output": pipeline_output,
            "diagram": mermaid_graph,
        }

    def _clear_visits_count(self):
        """
        Make sure all nodes's visits count is zero.
        """
        for node in self.graph.nodes:
            self.graph.nodes[node]["visits"] = 0

    def _check_max_loops(self, component_name: str):
        """
        Verify whether this component run too many times.
        """
        if self.graph.nodes[component_name]["visits"] > self.max_loops_allowed:
            raise PipelineMaxLoops(
                f"Maximum loops count ({self.max_loops_allowed}) exceeded for component '{component_name}'."
            )

    # This function is complex so it contains quite some logic, it needs tons of information
    # regarding a component to understand what action it should take so we have many local
    # variables and to keep things simple we also have multiple returns.
    # In the end this amount of information makes it easier to understand the internal logic so
    # we chose to ignore these pylint warnings.
    def _calculate_action(  # pylint: disable=too-many-locals, too-many-return-statements
        self, name: str, inputs: Dict[str, Any], inputs_buffer: Dict[str, Any]
    ) -> Literal["run", "wait", "skip", "remove"]:
        """
        Calculates the action to take for the component specified by `name`.
        There are four possible actions:
            * run
            * wait
            * skip
            * remove

        The below conditions are evaluated in this order.

        Component will run if at least one of the following statements is true:
            * It received all mandatory inputs
            * It received all mandatory inputs and it has no optional inputs
            * It received all mandatory inputs and all optional inputs are skipped
            * It received all mandatory inputs and some optional inputs and the rest are skipped
            * It received some of its inputs and the others are defaulted
            * It's the first component of the pipeline

        Component will wait if:
            * It received some of its inputs and the other are not skipped
            * It received all mandatory inputs and some optional inputs have not been skipped

        Component will be skipped if:
            * It never ran nor waited

        Component will be removed if:
            * It ran or waited at least once but can't do it again

        If none of the above condition is met a PipelineRuntimeError is raised.

        For simplicity sake input components that create a cycle, or components that already ran
        and don't create a cycle are considered as skipped.

        Args:
            name: Name of the component
            inputs: Values that the component will take as input
            inputs_buffer: Other components' inputs

        Returns:
            Action to take for component specifing whether it should run, wait, skip or be removed

        Raises:
            PipelineRuntimeError: If action to take can't be determined
        """

        # Upstream components/socket pairs the current component is connected to
        input_components = {
            from_node: data["to_socket"].name for from_node, _, data in self.graph.in_edges(name, data=True)
        }
        # Sockets that have received inputs from upstream components
        received_input_sockets = set(inputs.keys())

        # All components inputs, whether they're connected, default or pipeline inputs
        instance = self.graph.nodes[name]["instance"]
        input_sockets = {f.name for f in fields(instance.__canals_input__)}
        optional_input_sockets = set(instance.__canals_optional_inputs__)
        mandatory_input_sockets = set(instance.__canals_mandatory_inputs__)

        # Components that are in the inputs buffer and have no inputs assigned are considered skipped
        skipped_components = {n for n, v in inputs_buffer.items() if not v}

        # Sockets that have their upstream component marked as skipped
        skipped_optional_input_sockets = {
            sockets["to_socket"].name
            for from_node, _, sockets in self.graph.in_edges(name, data=True)
            if from_node in skipped_components and sockets["to_socket"].name in optional_input_sockets
        }

        for from_node, socket in input_components.items():
            if socket not in optional_input_sockets:
                continue
            loops_back = networkx.has_path(self.graph, name, from_node)
            has_run = self.graph.nodes[from_node]["visits"] > 0
            if loops_back or has_run:
                # Consider all input components that loop back to current component
                # or that have already run at least once as skipped.
                # This must be done to correctly handle cycles in the pipeline or we
                # would reach a dead lock in components that have multiple inputs and
                # one of these forms a cycle.
                skipped_optional_input_sockets.add(socket)

        ##############
        # RUN CHECKS #
        ##############
        if (
            mandatory_input_sockets.issubset(received_input_sockets)
            and input_sockets == received_input_sockets | mandatory_input_sockets | skipped_optional_input_sockets
        ):
            # We received all mandatory inputs and:
            #   * There are no optional inputs or
            #   * All optional inputs are skipped or
            #   * We received part of the optional inputs, the rest are skipped
            if not optional_input_sockets:
                logger.debug("Component '%s' is ready to run. All mandatory inputs received.", name)
            else:
                logger.debug(
                    "Component '%s' is ready to run. All mandatory inputs received, skipped optional inputs: %s",
                    name,
                    skipped_optional_input_sockets,
                )
            return "run"

        if set(input_components.values()).issubset(received_input_sockets):
            # We have data from each connected input component.
            # We reach this when the current component is the first of the pipeline or
            # when it has defaults and all its input components have run.
            logger.debug("Component '%s' is ready to run. All expected inputs were received.", name)
            return "run"

        ###############
        # WAIT CHECKS #
        ###############
        if mandatory_input_sockets == received_input_sockets and skipped_optional_input_sockets.issubset(
            optional_input_sockets
        ):
            # We received all of the inputs we need, but some optional inputs have not been run or skipped yet
            logger.debug(
                "Component '%s' is waiting. All mandatory inputs received, some optional are not skipped: %s",
                name,
                optional_input_sockets - skipped_optional_input_sockets,
            )
            return "wait"

        if any(self.graph.nodes[n]["visits"] == 0 for n in input_components.keys()):
            # Some upstream component that must send input to the current component has yet to run.
            logger.debug(
                "Component '%s' is waiting. Missing inputs: %s",
                name,
                set(input_components.values()),
            )
            return "wait"

        ###############
        # SKIP CHECKS #
        ###############
        if self.graph.nodes[name]["visits"] == 0:
            # It's the first time visiting this component, if it can't run nor wait
            # it's fine skipping it at this point.
            logger.debug("Component '%s' is skipped. It can't run nor wait.", name)
            return "skip"

        #################
        # REMOVE CHECKS #
        #################
        if self.graph.nodes[name]["visits"] > 0:
            # This component has already been visited at least once. If it can't run nor wait
            # there is no reason to skip it again. So we it must be removed.
            logger.debug("Component '%s' is removed. It can't run, wait or skip.", name)
            return "remove"

        # Can't determine action to take
        raise PipelineRuntimeError(
            f"Can't determine Component '{name}' action. "
            f"Mandatory input sockets: {mandatory_input_sockets}, "
            f"optional input sockets: {optional_input_sockets}, "
            f"received input: {list(inputs.keys())}, "
            f"input components: {list(input_components.keys())}, "
            f"skipped components: {skipped_components}, "
            f"skipped optional inputs: {skipped_optional_input_sockets}."
            f"This is likely a Canals bug. Please open an issue at https://github.com/deepset-ai/canals.",
        )

    def _skip_downstream_unvisited_nodes(self, component: str, inputs_buffer: OrderedDict) -> OrderedDict:
        """
        When a component is skipped, put all downstream nodes in the inputs buffer too: the might be skipped too,
        unless they are merge nodes. They will be evaluated later by the pipeline execution loop.
        """
        downstream_nodes = [e[1] for e in self.graph.out_edges(component)]
        for downstream_node in downstream_nodes:
            if downstream_node in inputs_buffer:
                continue
            if self.graph.nodes[downstream_node]["visits"] == 0:
                # Skip downstream nodes only if they never been visited
                inputs_buffer[downstream_node] = {}
        return inputs_buffer

    def _run_component(self, name: str, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Once we're confident this component is ready to run, run it and collect the output.
        """
        self.graph.nodes[name]["visits"] += 1
        instance = self.graph.nodes[name]["instance"]
        try:
            logger.info("* Running %s (visits: %s)", name, self.graph.nodes[name]["visits"])
            logger.debug("   '%s' inputs: %s", name, inputs)

            # Optional fields are defaulted to None so creation of the input dataclass doesn't fail
            # cause we're missing some argument
            optionals = {field: None for field in instance.__canals_optional_inputs__}

            # Pass the inputs as kwargs after adding the component's own defaults to them
            inputs = {**optionals, **instance.defaults, **inputs}
            input_dataclass = instance.input(**inputs)

            output_dataclass = instance.run(input_dataclass)

            # Unwrap the output
            logger.debug("   '%s' outputs: %s\n", name, output_dataclass.__dict__)

        except Exception as e:
            raise PipelineRuntimeError(
                f"{name} raised '{e.__class__.__name__}: {e}' \nInputs: {inputs}\n\n"
                "See the stacktrace above for more information."
            ) from e

        return output_dataclass

    def _route_output(
        self,
        node_name: str,
        node_results: Dict[str, Any],
        inputs_buffer: OrderedDict,
    ) -> OrderedDict:
        """
        Distrubute the outputs of the component into the input buffer of downstream components.

        Returns the updated inputs buffer.
        """
        # This is not a terminal node: find out where the output goes, to which nodes and along which edge
        is_decision_node_for_loop = (
            any(networkx.has_path(self.graph, edge[1], node_name) for edge in self.graph.out_edges(node_name))
            and len(self.graph.out_edges(node_name)) > 1
        )
        for edge_data in self.graph.out_edges(node_name, data=True):
            to_socket = edge_data[2]["to_socket"]
            from_socket = edge_data[2]["from_socket"]
            target_node = edge_data[1]

            # If this is a decision node and a loop is involved, we add to the input buffer only the nodes
            # that received their expected output and we leave the others out of the queue.
            if is_decision_node_for_loop and getattr(node_results, from_socket.name) is None:
                if networkx.has_path(self.graph, target_node, node_name):
                    # In case we're choosing to leave a loop, do not put the loop's node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're leaving the loop.",
                        target_node,
                    )
                else:
                    # In case we're choosing to stay in a loop, do not put the external node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're staying in the loop.",
                        target_node,
                    )
            else:
                # In all other cases, populate the inputs buffer for all downstream nodes, setting None to any
                # edge that did not receive input.
                if target_node not in inputs_buffer:
                    inputs_buffer[target_node] = {}  # Create the buffer for the downstream node if it's not there yet

                value_to_route = getattr(node_results, from_socket.name, None)
                if value_to_route:
                    inputs_buffer[target_node][to_socket.name] = value_to_route

        return inputs_buffer



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import List

from dataclasses import make_dataclass


from canals.testing import BaseTestComponent
from canals.component import component


@component
class Repeat:
    """
    Repeats the input value on all outputs.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int

        return Input

    def __init__(self, outputs: List[str] = ["output_1", "output_2", "output_3"]):
        self.outputs = outputs
        self._output_type = make_dataclass("Output", fields=[(val, int, None) for val in outputs])

    @component.output  # type: ignore
    def output(self):
        return self._output_type

    def run(self, data):
        output_dataclass = self.output()
        for output in self.outputs:
            setattr(output_dataclass, output, data.value)
        return output_dataclass


class TestRepeat(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Repeat(), tmp_path)

    def test_saveload_outputs(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Repeat(outputs=["one", "two"]), tmp_path)

    def test_repeat_default(self):
        component = Repeat()
        results = component.run(component.input(value=10))
        assert results == component.output(output_1=10, output_2=10, output_3=10)
        assert component.init_parameters == {}

    def test_repeat_init(self):
        component = Repeat(outputs=["one", "two"])
        results = component.run(component.input(value=10))
        assert results == component.output(one=10, two=10)
        assert component.init_parameters == {"outputs": ["one", "two"]}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
import pytest

from canals.pipeline import Pipeline, PipelineMaxLoops
from test.sample_components import AddFixedValue, Threshold, MergeLoop

import logging

logging.basicConfig(level=logging.DEBUG)


def test_max_loops():
    add = AddFixedValue()
    threshold = Threshold(threshold=100)
    merge = MergeLoop(expected_type=int)

    pipe = Pipeline(max_loops_allowed=10)
    pipe.add_component("add", add)
    pipe.add_component("threshold", threshold)
    pipe.add_component("merge", merge)
    pipe.connect("threshold.below", "add.value")
    pipe.connect("add.value", "merge.value_1")
    pipe.connect("merge.value", "threshold.value")
    with pytest.raises(PipelineMaxLoops):
        pipe.run({"merge": merge.input(value_2=1)})



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
import inspect
from typing import Protocol, Union, List, Any, get_origin, get_args
from dataclasses import fields, Field
from functools import wraps

from canals.errors import ComponentError
from canals.component.input_output import Connection, _input, _output


logger = logging.getLogger(__name__)


# We ignore too-few-public-methods Pylint error as this is only meant to be
# the definition of the Component interface.
# A concrete Component will have more than method in any case.
class Component(Protocol):  # pylint: disable=too-few-public-methods
    """
    Abstract interface of a Component.
    This is only used by type checking tools.

    If you want to create a new Component use the @component decorator.
    """

    def run(self, data: Any) -> Any:
        """
        Takes the Component input and returns its output.
        Input and output dataclasses types must be defined in separate methods
        decorated with @component.input and @component.output respectively.

        We use Any both as data and return types since dataclasses don't have a specific type.
        """

    @property
    def __canals_input__(self) -> type:
        pass

    @property
    def __canals_output__(self) -> type:
        pass

    @property
    def __canals_optional_inputs__(self) -> List[str]:
        pass

    @property
    def __canals_mandatory_inputs__(self) -> List[str]:
        pass


class _Component:
    """
    Marks a class as a component. Any class decorated with `@component` can be used by a Pipeline.

    All components must follow the contract below. This docstring is the source of truth for components contract.

    ### `@component` decorator

    All component classes must be decorated with the `@component` decorator. This allows Canals to discover them.

    ### `@component.input`

    All components must decorate one single method with the `@component.input` decorator. This method must return a
    dataclass, which will be used as structure of the input of the component.

    For example, if the node is expecting a list of Documents, the fields of the returned dataclass should be
    `documents: List[Document]`. Note that you don't need to decorate the dataclass youself: `@component.input` will
    add the decorator for you.

    Here is an example of such method:

    ```python
    @component.input
    def input(self):
        class Input:
            value: int
            add: int

        return Input
    ```

    Defaults are allowed, as much as default factories and other dataclass properties.

    By default `@component.input` sets `None` as default for all fields, regardless of their definition: this gives you
    the possibility of passing a part of the input to the pipeline without defining every field of the component.
    For example, using the above definition, you can create an Input dataclass as:

    ```python
    self.input(add=3)
    ```

    and the resulting dataclass will look like `Input(value=None, add=3)`.

    However, if you don't explicitly define them as Optionals, Pipeline will make sure to collect all the values of
    this dataclass before calling the `run()` method, making them in practice non-optional.

    If you instead define a specific field as Optional in the dataclass, then Pipeline will **not** wait for them, and
    will run the component as soon as all the non-optional fields have received a value or, if all fields are optional,
    if at least one of them received it.

    This behavior allows Canals to define loops by not waiting on both incoming inputs of the entry component of the
    loop, and instead running as soon as at least one of them receives a value.

    ### `@component.output`

    All components must decorate one single method with the `@component.output` decorator. This method must return a
    dataclass, which will be used as structure of the output of the component.

    For example, if the node is producing a list of Documents, the fields of the returned dataclass should be
    `documents: List[Document]`. Note that you don't need to decorate the dataclass youself: `@component.output` will
    add the decorator for you.

    Here is an example of such method:

    ```python
    @component.output
    def output(self):
        class Output:
            value: int

        return Output
    ```

    Defaults are allowed, as much as default factories and other dataclass properties.

    ### `__init__(self, **kwargs)`

    Optional method.

    Components may have an `__init__` method where they define:

    - `self.defaults = {parameter_name: parameter_default_value, ...}`:
        All values defined here will be sent to the `run()` method when the Pipeline calls it.
        If any of these parameters is also receiving input from other components, those have precedence.
        This collection of values is supposed to replace the need for default values in `run()` and make them
        dynamically configurable. Keep in mind that only these defaults will count at runtime: defaults given to
        the `Input` dataclass (see above) will be ignored.

    - `self.init_parameters = {same parameters that the __init__ method received}`:
        In this dictionary you can store any state the components wish to be persisted when they are saved.
        These values will be given to the `__init__` method of a new instance when the pipeline is loaded.
        Note that by default the `@component` decorator saves the arguments automatically.
        However, if a component sets their own `init_parameters` manually in `__init__()`, that will be used instead.
        Note: all of the values contained here **must be JSON serializable**. Serialize them manually if needed.

    Components should take only "basic" Python types as parameters of their `__init__` function, or iterables and
    dictionaries containing only such values. Anything else (objects, functions, etc) will raise an exception at init
    time. If there's the need for such values, consider serializing them to a string.

    _(TODO explain how to use classes and functions in init. In the meantime see `test/components/test_accumulate.py`)_

    The `__init__` must be extrememly lightweight, because it's a frequent operation during the construction and
    validation of the pipeline. If a component has some heavy state to initialize (models, backends, etc...) refer to
    the `warm_up()` method.


    ### `warm_up(self)`

    Optional method.

    This method is called by Pipeline before the graph execution. Make sure to avoid double-initializations,
    because Pipeline will not keep track of which components it called `warm_up()` on.


    ### `run(self, data)`

    Mandatory method.

    This is the method where the main functionality of the component should be carried out. It's called by
    `Pipeline.run()`.

    When the component should run, Pipeline will call this method with an instance of the dataclass returned by the
    method decorated with `@component.input`. This dataclass contains:

    - all the input values coming from other components connected to it,
    - if any is missing, the corresponding value defined in `self.defaults`, if it exists.

    `run()` must return a single instance of the dataclass declared through the method decorated with
    `@component.output`.

    Args:
        class_: the class that Canals should use as a component.
        serializable: whether to check, at init time, if the component can be saved with
        `save_pipelines()`.

    Returns:
        A class that can be recognized as a component.

    Raises:
        ComponentError: if the class provided has no `run()` method or otherwise doesn't respect the component contract.
    """

    def __init__(self):
        self.registry = {}

    @property
    def input(self):
        """
        TODO: Documentation
        """
        return _input

    @property
    def output(self):
        """
        TODO: Documentation
        """
        return _output

    def _decorate(self, class_):
        # '__canals_component__' is used to distinguish components from regular classes.
        # Its value is set to the desired component name: normally it is the class name, but it can technically be customized.
        class_.__canals_component__ = class_.__name__

        # Find input and output properties
        (input_, output) = _find_input_output(class_)

        # Save the input and output properties so it's easier to find them when running the Component since we won't
        # need to search the exact property name each time
        class_.__canals_input__ = input_
        class_.__canals_output__ = output

        # Save optional inputs, optionals inputs are those fields for the __canals_input__ dataclass
        # that have an Optional type.
        # Those are necessary to implement Components that can run with partial input, this gives us
        # the possibility to have cycles in Pipelines.
        class_.__canals_optional_inputs__ = property(_optional_inputs)
        class_.__canals_mandatory_inputs__ = property(_mandatory_inputs)

        # Check that the run method respects all constraints
        _check_run_signature(class_)

        # Makes sure the self.defaults and self.init_parameters dictionaries are always present
        class_.init_parameters = {}
        class_.defaults = {}

        # Automatically registers all the init parameters in an instance attribute called `init_parameters`.
        class_.__init__ = _save_init_params(class_.__init__)

        if class_.__name__ in self.registry:
            logger.error(
                "Component %s is already registered. Previous imported from '%s', new imported from '%s'",
                class_.__name__,
                self.registry[class_.__name__],
                class_,
            )

        self.registry[class_.__name__] = class_
        logger.debug("Registered Component %s", class_)

        return class_

    def __call__(self, class_=None):
        if class_:
            return self._decorate(class_)

        return self._decorate


component = _Component()


def _find_input_output(class_):
    """
    Finds the input and the output definitions for class_ and returns them.

    There must be only a single definition of input and output for class_, if either
    none or more than one are found raise ConnectionError.
    """
    inputs_found = []
    outputs_found = []

    # Get all properties of class_
    properties = inspect.getmembers(class_, predicate=lambda m: isinstance(m, property))
    for _, prop in properties:
        if not hasattr(prop, "fget") and not hasattr(prop.fget, "__canals_connection__"):
            continue

        # Field __canals_connection__ is set by _input and _output decorators
        if prop.fget.__canals_connection__ == Connection.INPUT:
            inputs_found.append(prop)
        elif prop.fget.__canals_connection__ == Connection.OUTPUT:
            outputs_found.append(prop)

    if (in_len := len(inputs_found)) != 1:
        # Raise if we don't find only a single input definition
        if in_len == 0:
            raise ComponentError(
                f"No input definition found in Component {class_.__name__}. "
                "Create a method that returns a dataclass defining the input and "
                "decorate it with @component.input() to fix the error."
            )
        raise ComponentError(f"Multiple input definitions found for Component {class_.__name__}.")

    if (in_len := len(outputs_found)) != 1:
        # Raise if we don't find only a single output definition
        if in_len == 0:
            raise ComponentError(
                f"No output definition found in Component {class_.__name__}. "
                "Create a method that returns a dataclass defining the output and "
                "decorate it with @component.output() to fix the error."
            )
        raise ComponentError(f"Multiple output definitions found for Component {class_.__name__}.")

    return (inputs_found[0], outputs_found[0])


def _check_run_signature(class_):
    """
    Check that the component's run() method exists and respects all constraints
    """
    # Check for run()
    if not hasattr(class_, "run"):
        raise ComponentError(f"{class_.__name__} must have a 'run()' method. See the docs for more information.")
    run_signature = inspect.signature(class_.run)

    # run() must take a single input param
    if len(run_signature.parameters) != 2:
        raise ComponentError("run() must accept only a single parameter called 'data'.")

    # The input param must be called data
    if not "data" in run_signature.parameters:
        raise ComponentError("run() must accept a parameter called 'data'.")


def _is_optional(field: Field) -> bool:
    """
    Utility method that returns whether a field has an Optional type or not.
    """
    return get_origin(field.type) is Union and type(None) in get_args(field.type)


def _optional_inputs(self) -> List[str]:
    """
    Return all field names of self that have an Optional type.
    This is meant to be set as a property in a Component.
    """
    return [f.name for f in fields(self.__canals_input__) if _is_optional(f)]


def _mandatory_inputs(self) -> List[str]:
    """
    Return all field names of self that don't have an Optional type.
    This is meant to be set as a property in a Component.
    """
    return [f.name for f in fields(self.__canals_input__) if not _is_optional(f)]


def _save_init_params(init_func):
    """
    Decorator that saves the init parameters of a component in `self.init_parameters`
    """

    @wraps(init_func)
    def wrapper(self, *args, **kwargs):
        # Call the actual __init__ function with the arguments
        init_func(self, *args, **kwargs)

        # Collect and store all the init parameters, preserving whatever the components might have already added there
        self.init_parameters = {**kwargs, **self.init_parameters}

    return wrapper



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional


from canals.testing import BaseTestComponent
from canals.component import component


@component
class Threshold:
    """
    Redirects the value, unchanged, along a different connection whether the value is above
    or below the given threshold.

    Single input, double output decision component.

    :param threshold: the number to compare the input value against. This is also a parameter.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int
            threshold: int = 10

        return Input

    @component.output  # type: ignore
    def output(self):
        class Output:
            above: int
            below: int

        return Output

    def __init__(self, threshold: Optional[int] = None):
        """
        :param threshold: the number to compare the input value against.
        """
        if threshold:
            self.defaults = {"threshold": threshold}

    def run(self, data):
        if data.value < data.threshold:
            return self.output(above=None, below=data.value)
        return self.output(above=data.value, below=None)


class TestThreshold(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Threshold(), tmp_path)

    def test_saveload_threshold(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Threshold(threshold=3), tmp_path)

    def test_threshold(self):
        component = Threshold()

        results = component.run(component.input(value=5, threshold=10))
        assert results == component.output(above=None, below=5)

        results = component.run(component.input(value=15, threshold=10))
        assert results == component.output(above=15, below=None)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from canals.testing import BaseTestComponent
from canals.component import component


@component
class Subtract:
    """
    Compute the difference between two values.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            first_value: int
            second_value: int

        return Input

    @component.output  # type: ignore
    def output(self):
        class Output:
            difference: int

        return Output

    def run(self, data):
        """
        :param first_value: name of the connection carrying the value to subtract from.
        :param second_value: name of the connection carrying the value to subtract.
        """
        return self.output(difference=data.first_value - data.second_value)


class TestSubtract(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Subtract(), tmp_path)

    def test_subtract(self):
        component = Subtract()
        results = component.run(component.input(first_value=10, second_value=7))
        assert results == component.output(difference=3)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional
from dataclasses import make_dataclass, asdict, is_dataclass

from canals.testing import BaseTestComponent
from canals.component import component


@component
class Sum:
    """
    Sums the values of all the input connections together.
    """

    def __init__(self, inputs=["value_1"]) -> None:
        # mypy complains that we can't Optional is not a type, so we ignore the error
        # cause we consider this to be correct
        self._input = make_dataclass("Input", fields=[(f, Optional[int]) for f in inputs])  # type: ignore

    @component.input  # type: ignore
    def input(self):
        return self._input

    @component.output  # type: ignore
    def output(self):
        class Output:
            total: int

        return Output

    def run(self, data):
        values = []
        if is_dataclass(data):
            values = [n for n in asdict(data).values() if n]
        return self.output(total=sum(values))


class TestSum(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Sum(), tmp_path)

    def test_sum_no_values(self):
        component = Sum(inputs=[])
        results = component.run(component.input())
        assert results == component.output(total=0)
        assert component.init_parameters == {"inputs": []}

    def test_sum_one_value(self):
        component = Sum()
        results = component.run(component.input(10))
        assert results == component.output(total=10)
        assert component.init_parameters == {}

    def test_sum_few_values(self):
        component = Sum(inputs=["value_1", "value_2", "value_3"])
        results = component.run(component.input(10, 11, 12))
        assert results == component.output(total=33)
        assert component.init_parameters == {"inputs": ["value_1", "value_2", "value_3"]}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import builtins
from typing import List, Union, Optional
from dataclasses import make_dataclass, is_dataclass, asdict

from canals.component import component
from canals.testing import BaseTestComponent


@component
class MergeLoop:
    """
    Takes multiple inputs and returns the first one that is not None.
    """

    def __init__(self, expected_type: Union[type, str], inputs: List[str] = ["value_1", "value_2"]):
        if isinstance(expected_type, str):
            self.expected_type = getattr(builtins, expected_type)
        else:
            self.expected_type = expected_type
        self.init_parameters = {"expected_type": self.expected_type.__name__}
        # mypy complains that we can't Optional is not a type, so we ignore the error
        # cause we consider this to be correct
        self._input = make_dataclass("Input", fields=[(f, Optional[self.expected_type]) for f in inputs])  # type: ignore

    @component.input  # type: ignore
    def input(self):
        return self._input

    @component.output  # type: ignore
    def output(self):
        class Output:
            value: self.expected_type  # type: ignore

        return Output

    def run(self, data):
        """
        Takes some inputs and returns the first one that is not None.
        """
        values = []
        if is_dataclass(data):
            values = asdict(data).values()
        for v in values:
            if v is not None:
                return self.output(value=v)
        return self.output(value=None)


class TestMergeLoop(BaseTestComponent):
    def test_saveload_builtin_type(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(MergeLoop(expected_type=int), tmp_path)

    # TODO
    # def test_saveload_complex_type(self, tmp_path):
    #     self.assert_can_be_saved_and_loaded_in_pipeline(MergeLoop(expected_type=List[int]), tmp_path)

    # def test_saveload_object_type(self, tmp_path):
    #     class MyObject: ...
    #     self.assert_can_be_saved_and_loaded_in_pipeline(MergeLoop(expected_type=MyObject()), tmp_path)

    def test_merge_first(self):
        component = MergeLoop(expected_type=int, inputs=["in_1", "in_2"])
        results = component.run(component.input(5, None))
        assert results == component.output(value=5)

    def test_merge_second(self):
        component = MergeLoop(expected_type=int, inputs=["in_1", "in_2"])
        results = component.run(component.input(None, 5))
        assert results == component.output(value=5)

    def test_merge_nones(self):
        component = MergeLoop(expected_type=int, inputs=["in_1", "in_2", "in_3"])
        results = component.run(component.input(None, None, None))
        assert results == component.output(value=None)

    def test_merge_one(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input(1))
        assert results == component.output(value=1)

    def test_merge_one_none(self):
        component = MergeLoop(expected_type=int)
        results = component.run(component.input())
        assert results == component.output(value=None)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging
from enum import Enum
from dataclasses import fields, is_dataclass, dataclass, asdict, MISSING

logger = logging.getLogger(__name__)


def _make_fields_optional(class_: type):
    """
    Takes a dataclass definition and modifies its __init__ so that all fields have
    a default value set.
    If a field has a default factory use it to set the default value.
    If a field has neither a default factory or value default to None.
    """
    defaults = []
    for field in fields(class_):
        default = field.default
        if field.default is MISSING and field.default_factory is MISSING:
            default = None
        elif field.default is MISSING and field.default_factory is not MISSING:
            default = field.default_factory()
        defaults.append(default)
    # mypy complains we're accessing __init__ on an instance but it's not in reality.
    # class_ is a class definition and not an instance of it, so we're good.
    # Also only I/O dataclasses are meant to be passed to this function making it a bit safer.
    class_.__init__.__defaults__ = tuple(defaults)  # type: ignore


def _make_comparable(class_: type):
    """
    Overwrites the existing __eq__ method of class_ with a custom one.
    This is meant to be used only in I/O dataclasses, it takes into account
    whether the fields are marked as comparable or not.

    This is necessary since the automatically created __eq__ method in dataclasses
    also verifies the type of the class. That causes it to fail if the I/O dataclass
    is returned by a function.

    In here we don't compare the types of self and other but only their fields.
    """

    def comparator(self, other) -> bool:
        if not is_dataclass(other):
            return False

        fields_ = [f.name for f in fields(self) if f.compare]
        other_fields = [f.name for f in fields(other) if f.compare]
        if not len(fields_) == len(other_fields):
            return False

        self_dict, other_dict = asdict(self), asdict(other)
        for field in fields_:
            if not self_dict[field] == other_dict[field]:
                return False

        return True

    setattr(class_, "__eq__", comparator)


class Connection(Enum):
    INPUT = 1
    OUTPUT = 2


def _input(input_function=None):
    """
    Decorator to mark a method that returns a dataclass defining a Component's input.

    The decorated function becomes a property.
    """

    def decorator(function):
        def wrapper(self):
            class_ = function(self)
            # If the user didn't explicitly declare the returned class
            # as dataclass we do it out of convenience
            if not is_dataclass(class_):
                class_ = dataclass(class_)

            _make_comparable(class_)
            _make_fields_optional(class_)

            return class_

        # Magic field to ease some further checks, we set it in the wrapper
        # function so we access it like this <class>.<function>.fget.__canals_connection__
        wrapper.__canals_connection__ = Connection.INPUT

        # If we don't set the documentation explicitly the user wouldn't be able to access
        # since we make wrapper a property and not the original function.
        # This is not essential but a really nice to have.
        return property(fget=wrapper, doc=function.__doc__)

    # Check if we're called as @_input or @_input()
    if input_function:
        # Called with parens
        return decorator(input_function)

    # Called without parens
    return decorator


def _output(output_function=None):
    """
    Decorator to mark a method that returns a dataclass defining a Component's output.

    The decorated function becomes a property.
    """

    def decorator(function):
        def wrapper(self):
            class_ = function(self)
            if not is_dataclass(class_):
                class_ = dataclass(class_)
            _make_comparable(class_)
            return class_

        # Magic field to ease some further checks, we set it in the wrapper
        # function so we access it like this <class>.<function>.fget.__canals_connection__
        wrapper.__canals_connection__ = Connection.OUTPUT

        # If we don't set the documentation explicitly the user wouldn't be able to access
        # since we make wrapper a property and not the original function.
        # This is not essential but a really nice to have.
        return property(fget=wrapper, doc=function.__doc__)

    # Check if we're called as @_output or @_output()
    if output_function:
        # Called with parens
        return decorator(output_function)

    # Called without parens
    return decorator



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import *
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.sample_components import Accumulate, AddFixedValue, Threshold, Sum, MergeLoop

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    accumulator = Accumulate()
    merge_loop = MergeLoop(expected_type=int, inputs=["in_1", "in_2"])
    summer = Sum(inputs=["in_1", "in_2"])
    pipeline = Pipeline(max_loops_allowed=10)
    pipeline.add_component("merge", merge_loop)
    pipeline.add_component("sum", summer)
    pipeline.add_component("below_10", Threshold(threshold=10))
    pipeline.add_component("add_one", AddFixedValue(add=1))
    pipeline.add_component("counter", accumulator)
    pipeline.add_component("add_two", AddFixedValue(add=2))

    pipeline.connect("merge", "below_10")
    pipeline.connect("below_10.below", "add_one.value")
    pipeline.connect("add_one", "counter")
    pipeline.connect("counter", "merge.in_1")
    pipeline.connect("below_10.above", "add_two.value")
    pipeline.connect("add_two", "sum.in_1")

    pipeline.draw(tmp_path / "looping_and_merge_pipeline.png")

    results = pipeline.run(
        {"merge": merge_loop.input(in_2=8), "sum": summer.input(in_2=2)},
    )
    pprint(results)
    print("accumulate: ", accumulator.state)

    assert results == {"sum": summer.output(total=23)}
    assert accumulator.state == 19


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import make_dataclass

import pytest

from canals.testing import BaseTestComponent
from canals.component import component


@component
class Remainder:
    """
    Redirects the value, unchanged, along the connection corresponding to the remainder
    of a division. For example, if `divisor=3`, the value `5` would be sent along
    the second output connection.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int

        return Input

    def __init__(self, divisor: int = 2):
        if divisor == 0:
            raise ValueError("Can't divide by zero")
        self.divisor = divisor

        self._output_type = make_dataclass(
            "Output", fields=[(f"remainder_is_{val}", int, None) for val in range(divisor)]
        )

    @component.output  # type: ignore
    def output(self):
        return self._output_type

    def run(self, data):
        """
        :param value: the value to check the remainder of.
        """
        remainder = data.value % self.divisor
        output = self.output()
        setattr(output, f"remainder_is_{remainder}", data.value)
        return output


class TestRemainder(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Remainder(), tmp_path)

    def test_saveload_divisor(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Remainder(divisor=1), tmp_path)

    def test_remainder_default(self):
        component = Remainder()
        results = component.run(component.input(value=3))
        assert results == component.output(remainder_is_1=3)

    def test_remainder_with_divisor(self):
        component = Remainder(divisor=4)
        results = component.run(component.input(value=3))
        assert results == component.output(remainder_is_3=3)

    def test_remainder_zero(self):
        with pytest.raises(ValueError):
            Remainder(divisor=0)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Dict

import logging
from dataclasses import dataclass, fields


logger = logging.getLogger(__name__)


@dataclass
class OutputSocket:
    name: str
    type: type


@dataclass
class InputSocket:
    name: str
    type: type
    sender: Optional[str] = None


def find_input_sockets(component) -> Dict[str, InputSocket]:
    """
    Find a component's input sockets.
    """
    return {f.name: InputSocket(name=f.name, type=f.type) for f in fields(component.__canals_input__)}


def find_output_sockets(component) -> Dict[str, OutputSocket]:
    """
    Find a component's output sockets.
    """
    return {f.name: OutputSocket(name=f.name, type=f.type) for f in fields(component.__canals_output__)}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Union, List

from canals.testing import BaseTestComponent
from canals.component import component


@component
class Concatenate:
    """
    Concatenates two values
    """

    @component.input
    def input(self):
        class Input:
            first: Union[List[str], str]
            second: Union[List[str], str]

        return Input

    @component.output
    def output(self):
        class Output:
            value: List[str]

        return Output

    def run(self, data):
        if type(data.first) is str and type(data.second) is str:
            res = [data.first, data.second]
        elif type(data.first) is list and type(data.second) is list:
            res = data.first + data.second
        elif type(data.first) is list and type(data.second) is str:
            res = data.first + [data.second]
        elif type(data.first) is str and type(data.second) is list:
            res = [data.first] + data.second

        return self.output(res)


class TestConcatenate(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Concatenate(), tmp_path)

    def test_input_lists(self):
        component = Concatenate()
        res = component.run(component.input(["This"], ["That"]))
        assert res == component.output(["This", "That"])

    def test_input_strings(self):
        component = Concatenate()
        res = component.run(component.input("This", "That"))
        assert res == component.output(["This", "That"])

    def test_input_first_list_second_string(self):
        component = Concatenate()
        res = component.run(component.input(["This"], "That"))
        assert res == component.output(["This", "That"])

    def test_input_first_string_second_list(self):
        component = Concatenate()
        res = component.run(component.input("This", ["That"]))
        assert res == component.output(["This", "That"])



from typing import Optional

import pytest

from canals.pipeline import Pipeline, PipelineValidationError
from canals.pipeline.sockets import InputSocket, OutputSocket
from canals.pipeline.validation import find_pipeline_inputs, find_pipeline_outputs

from test.sample_components import Double, AddFixedValue, Sum, Parity


def test_find_pipeline_input_no_input():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    assert find_pipeline_inputs(pipe.graph) == {"comp1": [], "comp2": []}


def test_find_pipeline_input_one_input():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")

    assert find_pipeline_inputs(pipe.graph) == {
        "comp1": [InputSocket(name="value", type=int)],
        "comp2": [],
    }


def test_find_pipeline_input_two_inputs_same_component():
    pipe = Pipeline()
    pipe.add_component("comp1", AddFixedValue())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")

    assert find_pipeline_inputs(pipe.graph) == {
        "comp1": [
            InputSocket(name="value", type=int),
            InputSocket(name="add", type=int),
        ],
        "comp2": [],
    }


def test_find_pipeline_input_some_inputs_different_components():
    pipe = Pipeline()
    pipe.add_component("comp1", AddFixedValue())
    pipe.add_component("comp2", Double())
    pipe.add_component("comp3", AddFixedValue())
    pipe.connect("comp1", "comp3")
    pipe.connect("comp2", "comp3.add")

    assert find_pipeline_inputs(pipe.graph) == {
        "comp1": [
            InputSocket(name="value", type=int),
            InputSocket(name="add", type=int),
        ],
        "comp2": [InputSocket(name="value", type=int)],
        "comp3": [],
    }


def test_find_pipeline_variable_input_nodes_in_the_pipeline():
    pipe = Pipeline()
    pipe.add_component("comp1", AddFixedValue())
    pipe.add_component("comp2", Double())
    pipe.add_component("comp3", Sum(inputs=["in_1", "in_2"]))

    assert find_pipeline_inputs(pipe.graph) == {
        "comp1": [
            InputSocket(name="value", type=int),
            InputSocket(name="add", type=int),
        ],
        "comp2": [InputSocket(name="value", type=int)],
        "comp3": [InputSocket(name="in_1", type=Optional[int]), InputSocket(name="in_2", type=Optional[int])],
    }


def test_find_pipeline_output_no_output():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    assert find_pipeline_outputs(pipe.graph) == {}


def test_find_pipeline_output_one_output():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")

    assert find_pipeline_outputs(pipe.graph) == {"comp2": [OutputSocket(name="value", type=int)]}


def test_find_pipeline_some_outputs_same_component():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Parity())
    pipe.connect("comp1", "comp2")

    assert find_pipeline_outputs(pipe.graph) == {
        "comp2": [OutputSocket(name="even", type=int), OutputSocket(name="odd", type=int)]
    }


def test_find_pipeline_some_outputs_different_components():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Parity())
    pipe.add_component("comp3", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp1", "comp3")

    assert find_pipeline_outputs(pipe.graph) == {
        "comp2": [OutputSocket(name="even", type=int), OutputSocket(name="odd", type=int)],
        "comp3": [
            OutputSocket(name="value", type=int),
        ],
    }


def test_validate_pipeline_input_pipeline_with_no_inputs():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")
    with pytest.raises(PipelineValidationError, match="This pipeline has no inputs."):
        pipe.run({})


def test_validate_pipeline_input_unknown_component():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    with pytest.raises(ValueError, match="Pipeline received data for unknown component\(s\): test_component"):
        pipe.run({"test_component": Double().input(value=1)})


def test_validate_pipeline_input_all_necessary_input_is_present():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    with pytest.raises(ValueError, match="Missing input: comp1.value"):
        pipe.run({})


def test_validate_pipeline_input_all_necessary_input_is_present_considering_defaults():
    pipe = Pipeline()
    pipe.add_component("comp1", AddFixedValue())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.run({"comp1": AddFixedValue().input(value=1)})
    pipe.run({"comp1": AddFixedValue().input(value=1, add=1)})
    with pytest.raises(ValueError, match="Missing input: comp1.value"):
        pipe.run({"comp1": AddFixedValue().input(add=1)})


def test_validate_pipeline_input_only_expected_input_is_present():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    with pytest.raises(ValueError, match="The input value of comp2 is already sent by node comp1"):
        pipe.run({"comp1": Double().input(value=1), "comp2": Double().input(value=1)})


def test_validate_pipeline_input_only_expected_input_is_present_including_unknown_names():
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")

    with pytest.raises(ValueError, match="Component comp1 is not expecting any input value called add"):
        pipe.run({"comp1": AddFixedValue().input(value=1, add=3)})


def test_validate_pipeline_input_only_expected_input_is_present_and_defaults_dont_interfere():
    pipe = Pipeline()
    pipe.add_component("comp1", AddFixedValue(add=10))
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.run({"comp1": AddFixedValue().input(value=1, add=5)})



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0

from canals.component import component
from canals.testing import BaseTestComponent


@component
class Double:
    """
    Doubles the input value.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int

        return Input

    @component.output  # type: ignore
    def output(self):
        class Output:
            value: int

        return Output

    def run(self, data):
        """
        Doubles the input value
        """
        return self.output(value=data.value * 2)


class TestDouble(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Double(), tmp_path)

    def test_double_default(self):
        component = Double()
        results = component.run(component.input(value=10))
        assert results == component.output(value=20)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Tuple, Optional, Union, List, Any, get_args, get_origin

import logging
import itertools

from canals.errors import PipelineConnectError
from canals.pipeline.sockets import InputSocket, OutputSocket


logger = logging.getLogger(__name__)


def parse_connection_name(connection: str) -> Tuple[str, Optional[str]]:
    """
    Returns component-connection pairs from a connect_to/from string
    """
    if "." in connection:
        split_str = connection.split(".", maxsplit=1)
        return (split_str[0], split_str[1])
    return connection, None


def _types_are_compatible(sender, receiver):  # pylint: disable=too-many-return-statements
    """
    Checks whether the source type is equal or a subtype of the destination type. Used to validate pipeline connections.

    Note: this method has no pretense to perform proper type matching. It especially does not deal with aliasing of
    typing classes such as `List` or `Dict` to their runtime counterparts `list` and `dict`. It also does not deal well
    with "bare" types, so `List` is treated differently from `List[Any]`, even though they should be the same.

    Consider simplifying the typing of your components if you observe unexpected errors during component connection.
    """
    if sender == receiver or receiver is Any:
        return True

    if sender is Any:
        return False

    try:
        if issubclass(sender, receiver):
            return True
    except TypeError:  # typing classes can't be used with issubclass, so we deal with them below
        pass

    sender_origin = get_origin(sender)
    receiver_origin = get_origin(receiver)

    if sender_origin is not Union and receiver_origin is Union:
        return any(_types_are_compatible(sender, union_arg) for union_arg in get_args(receiver))

    if not sender_origin or not receiver_origin or sender_origin != receiver_origin:
        return False

    sender_args = get_args(sender)
    receiver_args = get_args(receiver)
    if len(sender_args) > len(receiver_args):
        return False

    return all(_types_are_compatible(*args) for args in zip(sender_args, receiver_args))


def find_unambiguous_connection(
    sender_node: str, receiver_node: str, sender_sockets: List[OutputSocket], receiver_sockets: List[InputSocket]
) -> Tuple[OutputSocket, InputSocket]:
    """
    Find one single possible connection between two lists of sockets.
    """
    # List all combinations of sockets that match by type
    possible_connections = [
        (sender_sock, receiver_sock)
        for sender_sock, receiver_sock in itertools.product(sender_sockets, receiver_sockets)
        if _types_are_compatible(sender_sock.type, receiver_sock.type)
    ]

    # No connections seem to be possible
    if not possible_connections:
        connections_status_str = _connections_status(
            sender_node=sender_node,
            sender_sockets=sender_sockets,
            receiver_node=receiver_node,
            receiver_sockets=receiver_sockets,
        )

        # Both sockets were specified: explain why the types don't match
        if len(sender_sockets) == len(receiver_sockets) and len(sender_sockets) == 1:
            raise PipelineConnectError(
                f"Cannot connect '{sender_node}.{sender_sockets[0].name}' with '{receiver_node}.{receiver_sockets[0].name}': "
                f"their declared input and output types do not match.\n{connections_status_str}"
            )

        # Not both sockets were specified: explain there's no possible match on any pair
        connections_status_str = _connections_status(
            sender_node=sender_node,
            sender_sockets=sender_sockets,
            receiver_node=receiver_node,
            receiver_sockets=receiver_sockets,
        )
        raise PipelineConnectError(
            f"Cannot connect '{sender_node}' with '{receiver_node}': "
            f"no matching connections available.\n{connections_status_str}"
        )

    # There's more than one possible connection
    if len(possible_connections) > 1:
        # Try to match by name
        name_matches = [
            (out_sock, in_sock) for out_sock, in_sock in possible_connections if in_sock.name == out_sock.name
        ]
        if len(name_matches) != 1:
            # TODO allow for multiple connections at once if there is no ambiguity?
            # TODO give priority to sockets that have no default values?
            connections_status_str = _connections_status(
                sender_node=sender_node,
                sender_sockets=sender_sockets,
                receiver_node=receiver_node,
                receiver_sockets=receiver_sockets,
            )
            raise PipelineConnectError(
                f"Cannot connect '{sender_node}' with '{receiver_node}': more than one connection is possible "
                "between these components. Please specify the connection name, like: "
                f"pipeline.connect('{sender_node}.{possible_connections[0][0].name}', "
                f"'{receiver_node}.{possible_connections[0][1].name}').\n{connections_status_str}"
            )

    return possible_connections[0]


def _connections_status(
    sender_node: str, receiver_node: str, sender_sockets: List[OutputSocket], receiver_sockets: List[InputSocket]
):
    """
    Lists the status of the sockets, for error messages.
    """
    sender_sockets_entries = []
    for sender_socket in sender_sockets:
        socket_types = get_socket_type_desc(sender_socket.type)
        sender_sockets_entries.append(f" - {sender_socket.name} ({socket_types})")
    sender_sockets_list = "\n".join(sender_sockets_entries)

    receiver_sockets_entries = []
    for receiver_socket in receiver_sockets:
        socket_types = get_socket_type_desc(receiver_socket.type)
        receiver_sockets_entries.append(
            f" - {receiver_socket.name} ({socket_types}), "
            f"{'sent by '+receiver_socket.sender if receiver_socket.sender else 'available'}"
        )
    receiver_sockets_list = "\n".join(receiver_sockets_entries)

    return f"'{sender_node}':\n{sender_sockets_list}\n'{receiver_node}':\n{receiver_sockets_list}"


def get_socket_type_desc(type_):
    """
    Assembles a readable representation of the type of a connection. Can handle primitive types, classes, and
    arbitrarily nested structures of types from the typing module.
    """
    # get_args returns something only if this type has subtypes, in which case it needs to be printed differently.
    args = get_args(type_)

    if not args:
        if isinstance(type_, type):
            if type_.__name__.startswith("typing."):
                return type_.__name__[len("typing.") :]
            return type_.__name__

        # Literals only accept instances, not classes, so we need to account for those.
        if isinstance(type_, str):
            return f"'{type_}'"  # Quote strings
        if str(type_).startswith("typing."):
            return str(type_)[len("typing.") :]
        return str(type_)

    # Python < 3.10 support
    if hasattr(type_, "_name"):
        type_name = type_._name  # pylint: disable=protected-access
        # Support for Optionals and Unions in Python < 3.10
        if not type_name:
            if type(None) in args:
                type_name = "Optional"
                args = [a for a in args if a is not type(None)]
            else:
                if not any(isinstance(a, type) for a in args):
                    type_name = "Literal"
                else:
                    type_name = "Union"
    else:
        type_name = type_.__name__

    subtypes = ", ".join([get_socket_type_desc(subtype) for subtype in args if subtype is not type(None)])
    return f"{type_name}[{subtypes}]"



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import List, Set, Sequence, Tuple, Dict, Mapping, Literal, Union, Optional, Any
from enum import Enum
import re
from pathlib import Path

import pytest

from canals.errors import PipelineConnectError
from canals.pipeline import Pipeline, PipelineConnectError
from canals.component import component
from canals.pipeline.sockets import find_input_sockets, find_output_sockets
from canals.pipeline.connections import find_unambiguous_connection, get_socket_type_desc

from test.sample_components import AddFixedValue, Greet
from test._helpers import make_component


class TestClass1:
    ...


class TestClass2:
    ...


class TestClass3(TestClass1):
    ...


class TestEnum(Enum):
    TEST1 = TestClass1
    TEST2 = TestClass2


@pytest.mark.parametrize(
    "from_type,to_type",
    [
        pytest.param(str, str, id="same-primitives"),
        pytest.param(str, Optional[str], id="receiving-primitive-is-optional"),
        pytest.param(str, Union[int, str], id="receiving-type-is-union-of-primitives"),
        pytest.param(Union[int, str], Union[int, str], id="identical-unions"),
        pytest.param(Union[int, str], Union[int, str, bool], id="receiving-union-is-superset-of-sender"),
        pytest.param(str, Any, id="primitive-to-any"),
        pytest.param(TestClass1, TestClass1, id="same-class"),
        pytest.param(TestClass1, Optional[TestClass1], id="receiving-class-is-optional"),
        pytest.param(TestClass1, TestClass1, id="class-to-any"),
        pytest.param(TestClass3, TestClass1, id="subclass-to-class"),
        pytest.param(TestClass1, Union[int, TestClass1], id="receiving-type-is-union-of-classes"),
        pytest.param(TestClass3, Union[int, TestClass1], id="receiving-type-is-union-of-superclasses"),
        pytest.param(List[int], List[int], id="same-lists"),
        pytest.param(List[int], Optional[List[int]], id="receiving-list-is-optional"),
        pytest.param(List[int], List[Any], id="list-of-primitive-to-list-of-any"),
        pytest.param(List[TestClass1], List[TestClass1], id="list-of-same-classes"),
        pytest.param(List[TestClass3], List[TestClass1], id="list-of-subclass-to-list-of-class"),
        pytest.param(List[TestClass1], List[Any], id="list-of-classes-to-list-of-any"),
        pytest.param(List[Set[Sequence[bool]]], List[Set[Sequence[bool]]], id="nested-sequences-of-same-primitives"),
        pytest.param(
            List[Set[Sequence[bool]]],
            List[Set[Sequence[Any]]],
            id="nested-sequences-of-primitives-to-nested-sequences-of-any",
        ),
        pytest.param(
            List[Set[Sequence[TestClass1]]], List[Set[Sequence[TestClass1]]], id="nested-sequences-of-same-classes"
        ),
        pytest.param(
            List[Set[Sequence[TestClass3]]],
            List[Set[Sequence[TestClass1]]],
            id="nested-sequences-of-subclasses-to-nested-sequences-of-classes",
        ),
        pytest.param(
            List[Set[Sequence[TestClass1]]],
            List[Set[Sequence[Any]]],
            id="nested-sequences-of-classes-to-nested-sequences-of-any",
        ),
        pytest.param(Dict[str, int], Dict[str, int], id="same-dicts-of-primitives"),
        pytest.param(Dict[str, int], Dict[Any, int], id="dict-of-primitives-to-dict-of-any-keys"),
        pytest.param(Dict[str, int], Dict[str, Any], id="dict-of-primitives-to-dict-of-any-values"),
        pytest.param(Dict[str, int], Dict[Any, Any], id="dict-of-primitives-to-dict-of-any-key-and-values"),
        pytest.param(Dict[str, TestClass1], Dict[str, TestClass1], id="same-dicts-of-classes-values"),
        pytest.param(Dict[str, TestClass3], Dict[str, TestClass1], id="dict-of-subclasses-to-dict-of-classes"),
        pytest.param(Dict[str, TestClass1], Dict[Any, TestClass1], id="dict-of-classes-to-dict-of-any-keys"),
        pytest.param(Dict[str, TestClass1], Dict[str, Any], id="dict-of-classes-to-dict-of-any-values"),
        pytest.param(Dict[str, TestClass1], Dict[Any, Any], id="dict-of-classes-to-dict-of-any-key-and-values"),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[str, Dict[str, int]]],
            id="nested-mappings-of-same-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[str, Dict[Any, int]]],
            id="nested-mapping-of-primitives-to-nested-mapping-of-any-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[Any, Dict[str, int]]],
            id="nested-mapping-of-primitives-to-nested-mapping-of-higher-level-any-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[str, Dict[str, Any]]],
            id="nested-mapping-of-primitives-to-nested-mapping-of-any-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[Any, Dict[Any, Any]]],
            id="nested-mapping-of-primitives-to-nested-mapping-of-any-keys-and-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            id="nested-mappings-of-same-classes",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass3]]],
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            id="nested-mapping-of-subclasses-to-nested-mapping-of-classes",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[str, Dict[Any, TestClass1]]],
            id="nested-mapping-of-classes-to-nested-mapping-of-any-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[Any, Dict[str, TestClass1]]],
            id="nested-mapping-of-classes-to-nested-mapping-of-higher-level-any-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[str, Dict[str, Any]]],
            id="nested-mapping-of-classes-to-nested-mapping-of-any-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[Any, Dict[Any, Any]]],
            id="nested-mapping-of-classes-to-nested-mapping-of-any-keys-and-values",
        ),
        pytest.param(
            Literal["a", "b", "c"],
            Literal["a", "b", "c"],
            id="same-primitive-literal",
        ),
        pytest.param(
            Literal[TestEnum.TEST1],
            Literal[TestEnum.TEST1],
            id="same-enum-literal",
        ),
        pytest.param(
            Tuple[Optional[Literal["a", "b", "c"]], Union[Path, Dict[int, TestClass1]]],
            Tuple[Optional[Literal["a", "b", "c"]], Union[Path, Dict[int, TestClass1]]],
            id="identical-deeply-nested-complex-type",
        ),
    ],
)
def test_connect_compatible_types(from_type, to_type):
    comp1 = make_component(output=from_type)
    comp2 = make_component(input=to_type)

    pipe = Pipeline()
    pipe.add_component("c1", comp1)
    pipe.add_component("c2", comp2)
    pipe.connect("c1", "c2")
    assert list(pipe.graph.edges) == [("c1", "c2", "value/value")]


@pytest.mark.parametrize(
    "from_type, to_type",
    [
        pytest.param(int, bool, id="different-primitives"),
        pytest.param(TestClass1, TestClass2, id="different-classes"),
        pytest.param(TestClass1, TestClass3, id="class-to-subclass"),
        pytest.param(Any, int, id="any-to-primitive"),
        pytest.param(Any, TestClass2, id="any-to-class"),
        pytest.param(Optional[str], str, id="sending-primitive-is-optional"),
        pytest.param(Optional[TestClass1], TestClass1, id="sending-class-is-optional"),
        pytest.param(Optional[List[int]], List[int], id="sending-list-is-optional"),
        pytest.param(Union[int, str], str, id="sending-type-is-union"),
        pytest.param(Union[int, str, bool], Union[int, str], id="sending-union-is-superset-of-receiver"),
        pytest.param(Union[int, bool], Union[int, str], id="partially-overlapping-unions-with-primitives"),
        pytest.param(Union[int, TestClass1], Union[int, TestClass2], id="partially-overlapping-unions-with-classes"),
        pytest.param(List[int], List[str], id="different-lists-of-primitives"),
        pytest.param(List[int], List, id="list-of-primitive-to-bare-list"),  # is "correct", but we don't support it
        pytest.param(List[int], list, id="list-of-primitive-to-list-object"),  # is "correct", but we don't support it
        pytest.param(List[TestClass1], List[TestClass2], id="different-lists-of-classes"),
        pytest.param(List[TestClass1], List[TestClass3], id="lists-of-classes-to-subclasses"),
        pytest.param(List[Any], List[str], id="list-of-any-to-list-of-primitives"),
        pytest.param(List[Any], List[TestClass2], id="list-of-any-to-list-of-classes"),
        pytest.param(
            List[Set[Sequence[str]]], List[Set[Sequence[bool]]], id="nested-sequences-of-different-primitives"
        ),
        pytest.param(
            List[Set[Sequence[str]]], Set[List[Sequence[str]]], id="different-nested-sequences-of-same-primitives"
        ),
        pytest.param(
            List[Set[Sequence[TestClass1]]], List[Set[Sequence[TestClass2]]], id="nested-sequences-of-different-classes"
        ),
        pytest.param(
            List[Set[Sequence[TestClass1]]],
            List[Set[Sequence[TestClass3]]],
            id="nested-sequences-of-classes-to-subclasses",
        ),
        pytest.param(
            List[Set[Sequence[TestClass1]]],
            Set[List[Sequence[TestClass1]]],
            id="different-nested-sequences-of-same-class",
        ),
        pytest.param(
            List[Set[Sequence[Any]]],
            List[Set[Sequence[bool]]],
            id="nested-list-of-Any-to-nested-list-of-primitives",
        ),
        pytest.param(
            List[Set[Sequence[Any]]],
            List[Set[Sequence[TestClass2]]],
            id="nested-list-of-Any-to-nested-list-of-classes",
        ),
        pytest.param(Dict[str, int], Dict[int, int], id="different-dict-of-primitive-keys"),
        pytest.param(Dict[str, int], Dict[str, bool], id="different-dict-of-primitive-values"),
        pytest.param(Dict[str, TestClass1], Dict[str, TestClass2], id="different-dict-of-class-values"),
        pytest.param(Dict[str, TestClass1], Dict[str, TestClass3], id="different-dict-of-class-to-subclass-values"),
        pytest.param(Dict[Any, int], Dict[int, int], id="dict-of-Any-keys-to-dict-of-primitives"),
        pytest.param(Dict[str, Any], Dict[int, int], id="dict-of-Any-values-to-dict-of-primitives"),
        pytest.param(Dict[str, Any], Dict[int, TestClass1], id="dict-of-Any-values-to-dict-of-classes"),
        pytest.param(Dict[Any, Any], Dict[int, int], id="dict-of-Any-keys-and-values-to-dict-of-primitives"),
        pytest.param(Dict[Any, Any], Dict[int, TestClass1], id="dict-of-Any-keys-and-values-to-dict-of-classes"),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Mapping[str, Dict[str, Dict[str, int]]],
            id="different-nested-mappings-of-same-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[str, Dict[int, int]]],
            id="same-nested-mappings-of-different-primitive-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[int, Dict[str, int]]],
            id="same-nested-mappings-of-different-higer-level-primitive-keys",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            Dict[str, Mapping[str, Dict[str, bool]]],
            id="same-nested-mappings-of-different-primitive-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[str, Dict[str, TestClass2]]],
            id="same-nested-mappings-of-different-class-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            Dict[str, Mapping[str, Dict[str, TestClass2]]],
            id="same-nested-mappings-of-class-to-subclass-values",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[Any, int]]],
            Dict[str, Mapping[str, Dict[str, int]]],
            id="nested-mapping-of-Any-keys-to-nested-mapping-of-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[Any, Dict[Any, int]]],
            Dict[str, Mapping[str, Dict[str, int]]],
            id="nested-mapping-of-higher-level-Any-keys-to-nested-mapping-of-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, Any]]],
            Dict[str, Mapping[str, Dict[str, int]]],
            id="nested-mapping-of-Any-values-to-nested-mapping-of-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, Any]]],
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            id="nested-mapping-of-Any-values-to-nested-mapping-of-classes",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[Any, Any]]],
            Dict[str, Mapping[str, Dict[str, int]]],
            id="nested-mapping-of-Any-keys-and-values-to-nested-mapping-of-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[str, Dict[Any, Any]]],
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            id="nested-mapping-of-Any-keys-and-values-to-nested-mapping-of-classes",
        ),
        pytest.param(
            Literal["a", "b", "c"],
            Literal["x", "y"],
            id="different-literal-of-same-primitive",
        ),
        pytest.param(
            Literal["a", "b", "c"],
            Literal["a", "b"],
            id="subset-literal",
        ),
        pytest.param(
            Literal[TestEnum.TEST1],
            Literal[TestEnum.TEST2],
            id="different-literal-of-same-enum",
        ),
        pytest.param(
            Tuple[Optional[Literal["a", "b", "c"]], Union[Path, Dict[int, TestClass1]]],
            Tuple[Literal["a", "b", "c"], Union[Path, Dict[int, TestClass1]]],
            id="deeply-nested-complex-type-is-compatible-but-cannot-be-checked",
        ),
    ],
)
def test_connect_non_compatible_types(from_type, to_type):
    comp1 = make_component(output=from_type)
    comp2 = make_component(input=to_type)

    pipe = Pipeline()
    pipe.add_component("c1", comp1)
    pipe.add_component("c2", comp2)

    with pytest.raises(
        PipelineConnectError,
        match="Cannot connect 'c1.value' with 'c2.value': their declared input and output types do not match.",
    ):
        pipe.connect("c1", "c2")


def test_connect_sender_component_does_not_exist():
    add_1 = AddFixedValue()
    add_2 = AddFixedValue()

    pipe = Pipeline()
    pipe.add_component("first", add_1)
    pipe.add_component("second", add_2)
    with pytest.raises(ValueError, match="Component named third not found in the pipeline"):
        pipe.connect("third", "second")


def test_connect_receiver_component_does_not_exist():
    add_1 = AddFixedValue()
    add_2 = AddFixedValue()

    pipe = Pipeline()
    pipe.add_component("first", add_1)
    pipe.add_component("second", add_2)
    with pytest.raises(ValueError, match="Component named third not found in the pipeline"):
        pipe.connect("first", "third")


def test_connect_sender_socket_does_not_exist():
    add_1 = AddFixedValue()
    add_2 = AddFixedValue()

    pipe = Pipeline()
    pipe.add_component("first", add_1)
    pipe.add_component("second", add_2)
    with pytest.raises(PipelineConnectError, match="first.wrong does not exist"):
        pipe.connect("first.wrong", "second")


def test_connect_receiver_socket_does_not_exist():
    add_1 = AddFixedValue()
    add_2 = AddFixedValue()

    pipe = Pipeline()
    pipe.add_component("first", add_1)
    pipe.add_component("second", add_2)
    with pytest.raises(PipelineConnectError, match="second.wrong does not exist"):
        pipe.connect("first", "second.wrong")


def test_connect_many_outputs_to_the_same_input():
    add_1 = AddFixedValue()
    add_2 = AddFixedValue()

    pipe = Pipeline()
    pipe.add_component("first", add_1)
    pipe.add_component("second", add_2)
    pipe.add_component("third", add_2)
    pipe.connect("first.value", "second.value")
    with pytest.raises(PipelineConnectError, match="second.value is already connected to first"):
        pipe.connect("third.value", "second.value")


def test_connect_many_connections_possible_name_matches():
    @component
    class Component1:
        @component.input
        def input(self):
            class Input:
                value: str

            return Input

        @component.output
        def output(self):
            class Output:
                value: str

            return Output

        def run(self, data):
            return self.output(value=data.value)

    @component
    class Component2:
        @component.input
        def input(self):
            class Input:
                value: str
                othervalue: str
                yetanothervalue: str

            return Input

        @component.output
        def output(self):
            class Output:
                value: str

            return Output

        def run(self, data):
            return self.output(value=data.value)

    pipe = Pipeline()
    pipe.add_component("c1", Component1())
    pipe.add_component("c2", Component2())
    pipe.connect("c1", "c2")
    assert list(pipe.graph.edges) == [("c1", "c2", "value/value")]


def test_find_unambiguous_connection_many_connections_possible_no_name_matches():
    @component
    class Component1:
        @component.input
        def input(self):
            class Input:
                value: str

            return Input

        @component.output
        def output(self):
            class Output:
                value: str

            return Output

        def run(self, data):
            return self.output(value=data.value)

    @component
    class Component2:
        @component.input
        def input(self):
            class Input:
                value1: str
                value2: str
                value3: str

            return Input

        @component.output
        def output(self):
            class Output:
                value: str

            return Output

        def run(self, data):
            return self.output(value=data.value1)

    expected_message = re.escape(
        """Cannot connect 'comp1' with 'comp2': more than one connection is possible between these components. Please specify the connection name, like: pipeline.connect('comp1.value', 'comp2.value1').
'comp1':
 - value (str)
'comp2':
 - value1 (str), available
 - value2 (str), available
 - value3 (str), available"""
    )
    with pytest.raises(PipelineConnectError, match=expected_message):
        find_unambiguous_connection(
            sender_node="comp1",
            receiver_node="comp2",
            sender_sockets=find_output_sockets(Component1()).values(),
            receiver_sockets=find_input_sockets(Component2()).values(),
        )


@pytest.mark.parametrize(
    "type_,repr",
    [
        pytest.param(str, "str", id="primitive-types"),
        pytest.param(Any, "Any", id="any"),
        pytest.param(TestClass1, "TestClass1", id="class"),
        pytest.param(Optional[int], "Optional[int]", id="shallow-optional-with-primitive"),
        pytest.param(Optional[Any], "Optional[Any]", id="shallow-optional-with-any"),
        pytest.param(Optional[TestClass1], "Optional[TestClass1]", id="shallow-optional-with-class"),
        pytest.param(Union[bool, TestClass1], "Union[bool, TestClass1]", id="shallow-union"),
        pytest.param(List[str], "List[str]", id="shallow-sequence-of-primitives"),
        pytest.param(List[Set[Sequence[str]]], "List[Set[Sequence[str]]]", id="nested-sequence-of-primitives"),
        pytest.param(
            Optional[List[Set[Sequence[str]]]],
            "Optional[List[Set[Sequence[str]]]]",
            id="optional-nested-sequence-of-primitives",
        ),
        pytest.param(
            List[Set[Sequence[Optional[str]]]],
            "List[Set[Sequence[Optional[str]]]]",
            id="nested-optional-sequence-of-primitives",
        ),
        pytest.param(List[TestClass1], "List[TestClass1]", id="shallow-sequence-of-classes"),
        pytest.param(
            List[Set[Sequence[TestClass1]]], "List[Set[Sequence[TestClass1]]]", id="nested-sequence-of-classes"
        ),
        pytest.param(Dict[str, int], "Dict[str, int]", id="shallow-mapping-of-primitives"),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, int]]],
            "Dict[str, Mapping[str, Dict[str, int]]]",
            id="nested-mapping-of-primitives",
        ),
        pytest.param(
            Dict[str, Mapping[Any, Dict[str, int]]],
            "Dict[str, Mapping[Any, Dict[str, int]]]",
            id="nested-mapping-of-primitives-with-any",
        ),
        pytest.param(Dict[str, TestClass1], "Dict[str, TestClass1]", id="shallow-mapping-of-classes"),
        pytest.param(
            Dict[str, Mapping[str, Dict[str, TestClass1]]],
            "Dict[str, Mapping[str, Dict[str, TestClass1]]]",
            id="nested-mapping-of-classes",
        ),
        pytest.param(
            Literal["a", "b", "c"],
            "Literal['a', 'b', 'c']",
            id="string-literal",
        ),
        pytest.param(
            Literal[1, 2, 3],
            "Literal[1, 2, 3]",
            id="primitive-literal",
        ),
        pytest.param(
            Literal[TestEnum.TEST1],
            "Literal[TestEnum.TEST1]",
            id="enum-literal",
        ),
        pytest.param(
            Tuple[Optional[Literal["a", "b", "c"]], Union[Path, Dict[int, TestClass1]]],
            "Tuple[Optional[Literal['a', 'b', 'c']], Union[Path, Dict[int, TestClass1]]]",
            id="deeply-nested-complex-type",
        ),
    ],
)
def test_get_socket_type_desc(type_, repr):
    assert get_socket_type_desc(type_) == repr



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Any
from canals.component import component


def make_component(input=Any, output=Any):
    @component
    class Component:
        @component.input
        def input(self):
            class Input:
                value: input

            return Input

        @component.output
        def output(self):
            class Output:
                value: output

            return Output

        def run(self, data):
            return self.output()

    return Component()



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import os
import sys
import filecmp

from unittest.mock import patch, MagicMock
import pytest
import requests

from canals.pipeline import Pipeline
from canals.errors import PipelineDrawingError
from canals.draw import draw, convert

from test.sample_components import Double


@pytest.mark.skipif(sys.platform.lower().startswith("darwin"), reason="the available graphviz version is too recent")
@pytest.mark.skipif(sys.platform.lower().startswith("win"), reason="pygraphviz is not really available in Windows")
def test_draw_pygraphviz(tmp_path, test_files):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")

    draw(pipe.graph, tmp_path / "test_pipe.jpg", engine="graphviz")
    assert os.path.exists(tmp_path / "test_pipe.jpg")
    assert filecmp.cmp(tmp_path / "test_pipe.jpg", test_files / "pipeline_draw" / "pygraphviz.jpg")


def test_draw_mermaid_img(tmp_path, test_files):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    draw(pipe.graph, tmp_path / "test_pipe.jpg", engine="mermaid-img")
    assert os.path.exists(tmp_path / "test_pipe.jpg")
    assert filecmp.cmp(tmp_path / "test_pipe.jpg", test_files / "mermaid_mock" / "test_response.png")


def test_draw_mermaid_img_failing_request(tmp_path):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    with patch("canals.draw.mermaid.requests.get") as mock_get:

        def raise_for_status(self):
            raise requests.HTTPError()

        mock_response = MagicMock()
        mock_response.status_code = 429
        mock_response.content = '{"error": "too many requests"}'
        mock_response.raise_for_status = raise_for_status
        mock_get.return_value = mock_response

        with pytest.raises(PipelineDrawingError, match="There was an issue with https://mermaid.ink/"):
            draw(pipe.graph, tmp_path / "test_pipe.jpg", engine="mermaid-img")


def test_draw_mermaid_txt(tmp_path):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    draw(pipe.graph, tmp_path / "test_pipe.md", engine="mermaid-text")
    assert os.path.exists(tmp_path / "test_pipe.md")
    assert (
        open(tmp_path / "test_pipe.md", "r").read()
        == """graph TD;
comp1 -- value -> value --> comp2
comp2 -- value -> value --> comp1"""
    )


def test_draw_unknown_engine(tmp_path):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    with pytest.raises(ValueError, match="Unknown rendering engine 'unknown'"):
        draw(pipe.graph, tmp_path / "test_pipe.jpg", engine="unknown")


def test_convert_unknown_engine(tmp_path):
    pipe = Pipeline()
    pipe.add_component("comp1", Double())
    pipe.add_component("comp2", Double())
    pipe.connect("comp1", "comp2")
    pipe.connect("comp2", "comp1")

    with pytest.raises(ValueError, match="Unknown rendering engine 'unknown'"):
        convert(pipe.graph, engine="unknown")



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional


from canals.component import component
from canals.testing.test_component import BaseTestComponent


@component
class AddFixedValue:
    """
    Adds the value of `add` to `value`. If not given, `add` defaults to 1.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int
            add: int

        return Input

    @component.output  # type: ignore
    def output(self):
        class Output:
            value: int

        return Output

    def __init__(self, add: Optional[int] = 1):
        if add:
            self.defaults = {"add": add}

    def run(self, data):
        return self.output(value=data.value + data.add)


class TestAddFixedValue(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(AddFixedValue(), tmp_path)

    def test_saveload_add(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(AddFixedValue(add=2), tmp_path)

    def test_addvalue(self):
        component = AddFixedValue()
        results = component.run(component.input(value=50, add=10))
        assert results == component.output(value=60)
        assert component.init_parameters == {}



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import logging

import networkx

from networkx.drawing.nx_agraph import to_agraph as nx_to_agraph


logger = logging.getLogger(__name__)


def to_agraph(graph: networkx.MultiDiGraph):
    """
    Renders a pipeline graph using PyGraphViz. You need to install it and all its system dependencies for it to work.
    """
    try:
        import pygraphviz  # pylint: disable=unused-import,import-outside-toplevel
    except (ModuleNotFoundError, ImportError) as exc:
        raise ImportError(
            "Can't use 'pygraphviz' to draw this pipeline: pygraphviz could not be imported. "
            "Make sure pygraphviz is installed and all its system dependencies are setup correctly."
        ) from exc

    for inp, outp, key, data in graph.out_edges("input", keys=True, data=True):
        data["style"] = "dashed"
        graph.add_edge(inp, outp, key=key, **data)

    for inp, outp, key, data in graph.in_edges("output", keys=True, data=True):
        data["style"] = "dashed"
        graph.add_edge(inp, outp, key=key, **data)

    graph.nodes["input"]["shape"] = "plain"
    graph.nodes["output"]["shape"] = "plain"
    agraph = nx_to_agraph(graph)
    agraph.layout("dot")
    return agraph



from pathlib import Path

import pytest

from unittest.mock import patch, MagicMock


@pytest.fixture
def test_files():
    return Path(__file__).parent / "test_files"


@pytest.fixture(autouse=True)
def mock_mermaid_request(test_files):
    """
    Prevents real requests to https://mermaid.ink/
    """
    with patch("canals.draw.mermaid.requests.get") as mock_get:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = open(test_files / "mermaid_mock" / "test_response.png", "rb").read()
        mock_get.return_value = mock_response
        yield



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional


from canals.testing import BaseTestComponent
from canals.component import component


@component
class Parity:
    """
    Redirects the value, unchanged, along the 'even' connection if even, or along the 'odd' one if odd.
    """

    @component.input  # type: ignore
    def input(self):
        class Input:
            value: int

        return Input

    @component.output  # type: ignore
    def output(self):
        class Output:
            even: int = None
            odd: int = None

        return Output

    def run(self, data):
        """
        :param value: The value to check for parity
        """
        remainder = data.value % 2
        if remainder:
            return self.output(odd=data.value)
        return self.output(even=data.value)


class TestParity(BaseTestComponent):
    def test_saveload_default(self, tmp_path):
        self.assert_can_be_saved_and_loaded_in_pipeline(Parity(), tmp_path)

    def test_parity(self):
        component = Parity()
        results = component.run(component.input(value=1))
        assert results == component.output(odd=1)
        results = component.run(component.input(value=2))
        assert results == component.output(even=2)



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
import typing
from typing import List, Optional, Union, Set, Sequence, Iterable, Dict, Mapping, Tuple

from canals.component import component
from canals.pipeline.sockets import (
    find_input_sockets,
    InputSocket,
)


def test_find_input_sockets_one_regular_builtin_type_input():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: int

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=data.input_value)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {"input_value": InputSocket(name="input_value", type=int)}
    assert sockets == expected


def test_find_input_sockets_many_regular_builtin_type_inputs():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                int_value: int
                str_value: str
                bool_value: bool

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=data.int_value)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "int_value": InputSocket(name="int_value", type=int),
        "str_value": InputSocket(name="str_value", type=str),
        "bool_value": InputSocket(name="bool_value", type=bool),
    }
    assert sockets == expected


def test_find_input_sockets_one_regular_object_type_input():
    class MyObject:
        ...

    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: MyObject

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {"input_value": InputSocket(name="input_value", type=MyObject)}
    assert sockets == expected


def test_find_input_sockets_one_union_type_input():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: Union[str, int]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {"input_value": InputSocket(name="input_value", type=Union[str, int])}
    assert sockets == expected


def test_find_input_sockets_one_optional_builtin_type_input():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: Optional[int] = None

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {"input_value": InputSocket(name="input_value", type=Optional[int])}
    assert sockets == expected


def test_find_input_sockets_one_optional_object_type_input():
    class MyObject:
        ...

    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                input_value: Optional[MyObject] = None

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {"input_value": InputSocket(name="input_value", type=Optional[MyObject])}
    assert sockets == expected


def test_find_input_sockets_sequences_of_builtin_type_input():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                list_value: List[int]
                set_value: Set[int]
                sequence_value: Sequence[int]
                iterable_value: Iterable[int]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "list_value": InputSocket(name="list_value", type=typing.List[int]),
        "set_value": InputSocket(name="set_value", type=typing.Set[int]),
        "sequence_value": InputSocket(name="sequence_value", type=typing.Sequence[int]),
        "iterable_value": InputSocket(name="iterable_value", type=typing.Iterable[int]),
    }
    assert sockets == expected


def test_find_input_sockets_sequences_of_object_type_input():
    class MyObject:
        ...

    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                list_value: List[MyObject]
                set_value: Set[MyObject]
                sequence_value: Sequence[MyObject]
                iterable_value: Iterable[MyObject]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "list_value": InputSocket(name="list_value", type=typing.List[MyObject]),
        "set_value": InputSocket(name="set_value", type=typing.Set[MyObject]),
        "sequence_value": InputSocket(name="sequence_value", type=typing.Sequence[MyObject]),
        "iterable_value": InputSocket(name="iterable_value", type=typing.Iterable[MyObject]),
    }
    assert sockets == expected


def test_find_input_sockets_mappings_of_builtin_type_input():
    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                dict_value: Dict[str, int]
                mapping_value: Mapping[str, int]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "dict_value": InputSocket(name="dict_value", type=typing.Dict[str, int]),
        "mapping_value": InputSocket(name="mapping_value", type=typing.Mapping[str, int]),
    }
    assert sockets == expected


def test_find_input_sockets_mappings_of_object_type_input():
    class MyObject:
        ...

    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                dict_value: Dict[str, MyObject]
                mapping_value: Mapping[str, MyObject]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "dict_value": InputSocket(name="dict_value", type=typing.Dict[str, MyObject]),
        "mapping_value": InputSocket(name="mapping_value", type=typing.Mapping[str, MyObject]),
    }
    assert sockets == expected


def test_find_input_sockets_tuple_type_input():
    class MyObject:
        ...

    @component
    class MockComponent:
        @component.input
        def input(self):
            class Input:
                tuple_value: Tuple[str, MyObject]

            return Input

        @component.output
        def output(self):
            class Output:
                output_value: int

            return Output

        def run(self, data):
            return self.output(output_value=1)

    comp = MockComponent()
    sockets = find_input_sockets(comp)
    expected = {
        "tuple_value": InputSocket(name="tuple_value", type=typing.Tuple[str, MyObject]),
    }
    assert sockets == expected



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from typing import List, Dict, Any
import logging
from dataclasses import fields

import networkx

from canals.errors import PipelineValidationError
from canals.pipeline.sockets import InputSocket, OutputSocket


logger = logging.getLogger(__name__)


def find_pipeline_inputs(graph: networkx.MultiDiGraph) -> Dict[str, List[InputSocket]]:
    """
    Collect components that have disconnected input sockets. Note that this method returns *ALL* disconnected
    input sockets, including all such sockets with default values.
    """
    return {
        node: [socket for socket in data.get("input_sockets", {}).values() if not socket.sender]
        for node, data in graph.nodes(data=True)
    }


def find_pipeline_outputs(graph) -> Dict[str, List[OutputSocket]]:
    """
    Collect components that have disconnected output sockets. They define the pipeline output.
    """
    return {
        node: list(data.get("output_sockets", {}).values())
        for node, data in graph.nodes(data=True)
        if not graph.out_edges(node)
    }


def validate_pipeline_input(graph: networkx.MultiDiGraph, input_values: Dict[str, Any]) -> Dict[str, Any]:
    """
    Make sure the pipeline is properly built and that the input received makes sense.
    Returns the input values, validated and updated at need.
    """
    if not any(sockets for sockets in find_pipeline_inputs(graph).values()):
        raise PipelineValidationError("This pipeline has no inputs.")

    # Make sure the input keys are all nodes of the pipeline
    unknown_components = [key for key in input_values.keys() if not key in graph.nodes]
    if unknown_components:
        raise ValueError(f"Pipeline received data for unknown component(s): {', '.join(unknown_components)}")

    # Make sure all necessary sockets are connected
    _validate_input_sockets_are_connected(graph, input_values)

    # Make sure that the pipeline input is only sent to nodes that won't receive data from other nodes
    _validate_nodes_receive_only_expected_input(graph, input_values)

    return input_values


def _validate_input_sockets_are_connected(graph: networkx.MultiDiGraph, input_values: Dict[str, Any]):
    """
    Make sure all the inputs nodes are receiving all the values they need, either from the Pipeline's input or from
    other nodes.
    """
    valid_inputs = find_pipeline_inputs(graph)
    for node, sockets in valid_inputs.items():
        for socket in sockets:
            node_instance = graph.nodes[node]["instance"]
            input_in_node_defaults = hasattr(node_instance, "defaults") and socket.name in node_instance.defaults
            inputs_for_node = input_values.get(node)
            missing_input_value = (
                not inputs_for_node
                or not socket.name in [f.name for f in fields(inputs_for_node)]
                or not getattr(inputs_for_node, socket.name)
            )
            if missing_input_value and not input_in_node_defaults:
                raise ValueError(f"Missing input: {node}.{socket.name}")


def _validate_nodes_receive_only_expected_input(graph: networkx.MultiDiGraph, input_values: Dict[str, Any]):
    """
    Make sure that every input node is only receiving input values from EITHER the pipeline's input or another node,
    but never from both.
    """
    for node, input_data in input_values.items():
        for socket_name in [f.name for f in fields(input_data)]:
            if not getattr(input_data, socket_name):
                continue
            if not socket_name in graph.nodes[node]["input_sockets"].keys():
                raise ValueError(
                    f"Component {node} is not expecting any input value called {socket_name}. "
                    "Are you using the correct Input class?"
                )

            sender = graph.nodes[node]["input_sockets"][socket_name].sender
            if sender:
                raise ValueError(f"The input {socket_name} of {node} is already sent by node {sender}")



# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from canals.pipeline import Pipeline, save_pipelines, load_pipelines

#
# TODO right now the tests iterate on components, so it's not always clear which component fails the tests.
# Can we instead parametrize the tests on the output of the components fixture?
#
class BaseTestComponent:  # pylint: disable=too-few-public-methods
    """
    Base tests for components.
    """

    def assert_can_be_saved_and_loaded_in_pipeline(self, component, tmp_path):
        """
        Verifies whether the components can be properly serialized and de-serialized.
        """
        pipe = Pipeline()
        pipe.add_component("component", component)
        save_pipelines({"pipe": pipe}, tmp_path / "test_pipe.json")

        loaded_pipes = load_pipelines(tmp_path / "test_pipe.json")
        new_pipe = loaded_pipes["pipe"]

        assert pipe == new_pipe
